from typing import Optional
from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme, InspectorResult as Label
from nameguard.grapheme_normalization import grapheme_is_normalized
from nameguard.models import CheckStatus, Check, GraphemeCheckResult, LabelCheckResult, NameCheckResult


STATUS = CheckStatus.ALERT

MESSAGE_PASS = 'Valid for use with ENS'

MESSAGE_FAIL = 'Invalid for use with ENS'

L_MESSAGE_SKIP = 'This label is unknown'
N_MESSAGE_SKIP = 'This name contains unknown labels'


def check_grapheme(grapheme: Grapheme) -> GraphemeCheckResult:
    passed = grapheme_is_normalized(grapheme.value)
    return GraphemeCheckResult(
        check=Check.NORMALIZED,
        status=CheckStatus.PASS if passed else STATUS,
        _grapheme_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )


def check_label(label: Optional[Label]) -> LabelCheckResult:
    if label is None:
        # unknown label
        return LabelCheckResult(
            check=Check.NORMALIZED,
            status=CheckStatus.SKIP,
            _label_message=L_MESSAGE_SKIP,
            _name_message=N_MESSAGE_SKIP,
        )
    passed = label.status == 'normalized' and len(label.label) > 0
    return LabelCheckResult(
        check=Check.NORMALIZED,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )


def check_name(name: list[Optional[Label]]) -> NameCheckResult:
    if any(label is None for label in name):
        # unknown label in name
        return NameCheckResult(
            check=Check.NORMALIZED,
            status=CheckStatus.SKIP,
            _name_message=N_MESSAGE_SKIP,
        )
    passed = all(label.status == 'normalized' and len(label.label) > 0 for label in name)
    return NameCheckResult(
        check=Check.NORMALIZED,
        status=CheckStatus.PASS if passed else STATUS,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )



from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme
from nameguard.models import CheckStatus, Check, GenericCheckResult, GraphemeCheckResult


STATUS = CheckStatus.WARN

MESSAGE_PASS = 'Broadly accessible to type'

MESSAGE_FAIL = 'May be difficult to type on some devices'

def check_grapheme(grapheme: Grapheme) -> GenericCheckResult:
    passed = grapheme.type in (
        'simple_letter',
        'simple_number',
        'hyphen',
        'dollarsign',
        'underscore',
    )
    return GraphemeCheckResult(
        check=Check.TYPING_DIFFICULTY,
        status=CheckStatus.PASS if passed else STATUS,
        _grapheme_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )



from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme
from nameguard.models import CheckStatus, Check, GenericCheckResult, GraphemeCheckResult


STATUS = CheckStatus.ALERT

G_MESSAGE_PASS = 'Visible'
L_MESSAGE_PASS = 'All characters are visible'
N_MESSAGE_PASS = 'All characters are visible'

G_MESSAGE_FAIL = 'Invisible'
L_MESSAGE_FAIL = 'Contains invisible characters'
N_MESSAGE_FAIL = 'Contains invisible characters'


def check_grapheme(grapheme: Grapheme) -> GenericCheckResult:
    passed = grapheme.type != 'invisible'
    return GraphemeCheckResult(
        check=Check.INVISIBLE,
        status=CheckStatus.PASS if passed else STATUS,
        _grapheme_message=G_MESSAGE_PASS if passed else G_MESSAGE_FAIL,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
    )



from ens_normalize.normalization import NORMALIZATION
from ens_normalize import is_ens_normalized


def grapheme_is_normalized(conf: str) -> bool:
    if len(conf) == 1 and ord(conf) in NORMALIZATION.valid:
        return True
    return is_ens_normalized(conf)



from typing import Optional
from label_inspector.models import InspectorResult
from nameguard.models import CheckStatus, Check, GenericCheckResult, LabelCheckResult


# TODO: Should check if the *name* contains mixed scripts, not all labels


STATUS = CheckStatus.WARN

L_MESSAGE_PASS = 'Written in a single script'
N_MESSAGE_PASS = 'Written in a single script'

L_MESSAGE_FAIL = 'Written in multiple scripts'
N_MESSAGE_FAIL = 'Written in multiple scripts'

L_MESSAGE_SKIP = 'This label is unknown'
N_MESSAGE_SKIP = 'This name contains unknown labels'


def check_label(label: Optional[InspectorResult]) -> GenericCheckResult:
    if label is None:
        return LabelCheckResult(
            check=Check.MIXED_SCRIPTS,
            status=CheckStatus.SKIP,
            _label_message=L_MESSAGE_SKIP,
            _name_message=N_MESSAGE_SKIP,
        )
    passed = label.all_script is not None or label.label == ''
    return LabelCheckResult(
        check=Check.MIXED_SCRIPTS,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
    )



from typing import Optional
from label_inspector.models import InspectorResult
from nameguard.models import CheckStatus, Check, GenericCheckResult, LabelCheckResult


STATUS = CheckStatus.ALERT

L_MESSAGE_PASS =  'This label is known'
N_MESSAGE_PASS =  'This name is known'

L_MESSAGE_FAIL =  'This label is unknown'
N_MESSAGE_FAIL =  'This name contains unknown labels'


def check_label(label: Optional[InspectorResult]) -> GenericCheckResult:
    passed = label is not None
    return LabelCheckResult(
        check=Check.UNKNOWN_LABEL,
        status=CheckStatus.PASS if passed else STATUS,
        _label_message=L_MESSAGE_PASS if passed else L_MESSAGE_FAIL,
        _name_message=N_MESSAGE_PASS if passed else N_MESSAGE_FAIL,
    )



from typing import Optional
from nameguard.models import CheckStatus, Check, GenericCheckResult, NameCheckResult
from label_inspector.models import InspectorResult


MESSAGE_INFO = 'NameWrapper fuse checks will be added soon'


def check_name(labels: list[Optional[InspectorResult]]) -> GenericCheckResult:
    return NameCheckResult(
        check=Check.NAMEWRAPPER_FUSES,
        status=CheckStatus.INFO,
        _name_message=MESSAGE_INFO,
    )



from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme

from nameguard.grapheme_normalization import grapheme_is_normalized
from nameguard.models import CheckStatus, Check, GenericCheckResult, GraphemeCheckResult


STATUS = CheckStatus.WARN

MESSAGE_PASS = 'Unlikely to be confused'

MESSAGE_FAIL = 'May be confusable'

MESSAGE_SKIP = 'Confusable checks were skipped'

def check_grapheme(grapheme: Grapheme) -> GenericCheckResult:
    if not isinstance(grapheme, Grapheme) or not grapheme_is_normalized(grapheme.value):
        return GraphemeCheckResult(
            check=Check.CONFUSABLES,
            status=CheckStatus.SKIP,
            _grapheme_message=MESSAGE_SKIP,
            _label_message=MESSAGE_SKIP,
            _name_message=MESSAGE_SKIP,
        )

    passed = len(grapheme.confusables_other) == 0 and (grapheme.confusables_canonical is None or grapheme.confusables_canonical == grapheme.value)
    return GraphemeCheckResult(
        check=Check.CONFUSABLES,
        status=CheckStatus.PASS if passed else STATUS,
        _grapheme_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )



from fastapi import HTTPException
from pydantic import BaseModel, Field


class ExceptionResponse(BaseModel):
    detail: str = Field(
        description='Human-readable description of the error.',
        examples=['This is a human-readable description of the error.'],
    )


class NameGuardException(HTTPException):
    STATUS_CODE = None
    DESCRIPTION = None
    
    def __init__(self, detail: str = ''):
        super().__init__(self.STATUS_CODE, detail=f'{self.DESCRIPTION} {detail}'.strip())

    @classmethod
    def get_responses_spec(cls):
        return {cls.STATUS_CODE: {'description': cls.DESCRIPTION, 'model': ExceptionResponse}}


class InvalidNameHash(NameGuardException):
    STATUS_CODE = 422
    DESCRIPTION = "Provided namehash is not valid."

class InvalidTokenID(NameGuardException):
    STATUS_CODE = 422
    DESCRIPTION = "Provided token id is not valid."

class ENSSubgraphUnavailable(NameGuardException):
    STATUS_CODE = 503
    DESCRIPTION = "Error while making request to ENS Subgraph."


class NamehashMismatchError(NameGuardException):
    STATUS_CODE = 500
    DESCRIPTION = "Namehash calculated on the name returned from ENS Subgraph does not equal the input namehash."


class NamehashNotFoundInSubgraph(NameGuardException):
    STATUS_CODE = 404
    DESCRIPTION = "Provided namehash could not be found in ENS Subgraph."


class NotAGrapheme(NameGuardException):
    STATUS_CODE = 422
    DESCRIPTION = 'The provided string is not a single grapheme.'


class InvalidEthereumAddress(NameGuardException):
    STATUS_CODE = 422
    DESCRIPTION = "Provided Ethereum address is not valid."


class ProviderUnavailable(NameGuardException):
    STATUS_CODE = 503
    DESCRIPTION = "Error while making request to provider."


class MissingTitle(NameGuardException):
    STATUS_CODE = 422
    DESCRIPTION = "A field named 'title' is required within the list of fields"


import os
import re

from nameguard.our_ens import OurENS
from ens_normalize import ens_process, is_ens_normalized, ens_cure, DisallowedSequence

import requests
from label_inspector.inspector import Inspector
from label_inspector.config import initialize_inspector_config
from label_inspector.models import InspectorConfusableGraphemeResult
from web3 import HTTPProvider
from dotenv import load_dotenv

from nameguard import checks
from nameguard.models import (
    CheckStatus,
    Check,
    NameGuardReport,
    LabelGuardReport,
    ConsolidatedGraphemeGuardReport,
    BulkNameGuardBulkReport,
    Normalization,
    GraphemeNormalization,
    GraphemeGuardReport,
    NetworkName,
    SecurePrimaryNameResult,
    SecurePrimaryNameStatus,
    FakeEthNameCheckStatus,
    FakeEthNameCheckResult,
    CheckStatus,
    ImpersonationStatus,
)
from nameguard.provider import get_nft_metadata
from nameguard.utils import (
    namehash_from_name,
    labelhash_from_label,
    calculate_nameguard_rating,
    count_risks,
    agg_checks,
    get_highest_risk,
    label_is_labelhash,
    compute_canonical_from_list,
    is_labelhash_eth,
)
from nameguard.exceptions import ProviderUnavailable, NotAGrapheme, MissingTitle
from nameguard.logging import logger
from nameguard.subgraph import namehash_to_name_lookup, resolve_all_labelhashes_in_name, \
    resolve_all_labelhashes_in_name_querying_labelhashes, resolve_all_labelhashes_in_names_querying_labelhashes
from nameguard.generic_utils import capitalize_words

DNA_CHECKS = [
    (checks.dna.normalized.check_grapheme, checks.dna.normalized.check_label, checks.dna.normalized.check_name),
]

GRAPHEME_CHECKS = [
    checks.grapheme.confusables.check_grapheme,
    checks.grapheme.font_support.check_grapheme,
    checks.grapheme.invisible.check_grapheme,
    checks.grapheme.typing_difficulty.check_grapheme,
]

LABEL_CHECKS = [
    checks.label.mixed_scripts.check_label,
    checks.label.namewrapper.check_label,
    checks.label.punycode.check_label,
    checks.label.unknown.check_label,
]

NAME_CHECKS = [
    checks.name.impersonation_risk.check_name,
    checks.name.punycode_name.check_name,
    checks.name.namewrapper_fuses.check_name,
    checks.name.decentralized_name.check_name,
]


def init_inspector():
    with initialize_inspector_config('prod_config') as config:
        return Inspector(config)


ens_contract_adresses = {
    '0x57f1887a8bf19b14fc0df6fd9b2acc9af147ea85',  # Base Registrar
    '0xd4416b13d2b3a9abae7acd5d6c2bbdbe25686401',  # Name Wrapper
}

ALCHEMY_UNKNOWN_NAME = re.compile('^\[0x[0-9a-f]{4}\.\.\.[0-9a-f]{4}\]\.eth$')

def nested_get(dic, keys):
    for key in keys:
        dic = dic[key]
    return dic


class NameGuard:
    def __init__(self):
        self._inspector = init_inspector()
        load_dotenv()
        # TODO use web sockets and async
        self.ns = {}
        for network_name, env_var in ((NetworkName.MAINNET, 'PROVIDER_URI_MAINNET'),
                                      (NetworkName.GOERLI, 'PROVIDER_URI_GOERLI'),
                                      (NetworkName.SEPOLIA, 'PROVIDER_URI_SEPOLIA')):
            if os.environ.get(env_var) is None:
                logger.warning(f'Environment variable {env_var} is not set')
            self.ns[network_name] = OurENS(HTTPProvider(os.environ.get(env_var)))

    def analyse_label(self, label: str):
        return self._inspector.analyse_label(label, simple_confusables=True, omit_cure=True)

    async def inspect_name(self, network_name: NetworkName, name: str, resolve_labelhashes: bool = True) -> NameGuardReport:
        '''
        Inspect a name. A name is a sequence of labels separated by dots.
        A label can be a labelhash or a string.
        If a labelhash is encountered, it will be treated as an unknown label.
        '''

        logger.debug(f'[inspect_name] name: \'{name}\'')
        
        if resolve_labelhashes:
            name = await resolve_all_labelhashes_in_name_querying_labelhashes(network_name, name)

        labels = [] if len(name) == 0 else name.split('.')
        logger.debug(f'[inspect_name] labels: {labels}')

        # labelhashes have `None` as their analysis
        labels_analysis = [self.analyse_label(label)
                           # do not analyze labelhashes
                           if not label_is_labelhash(label)
                           else None
                           for label in labels]

        # -- check individual entities --

        # checks for each grapheme in each label
        labels_graphemes_checks = [
            [
                [check(grapheme) for check in GRAPHEME_CHECKS]
                for grapheme in label_analysis.graphemes
            ] if label_analysis is not None else []
            # label has [] graphemes if it's a labelhash
            for label_analysis in labels_analysis
        ]

        # checks for each label
        labels_checks = [
            [check(label_analysis) for check in LABEL_CHECKS]
            # checks have to handle labelhashes
            for label_analysis in labels_analysis
        ]

        # checks for the whole name
        name_checks = [check(labels_analysis) for check in NAME_CHECKS]

        # -- aggregate results --

        # merge grapheme checks into label checks
        for label_i, label_graphemes_checks in enumerate(labels_graphemes_checks):
            for grapheme_checks in label_graphemes_checks:
                labels_checks[label_i].extend([c.raise_context() for c in grapheme_checks])
            labels_checks[label_i] = agg_checks(labels_checks[label_i])

        # merge label checks into name checks
        for label_checks in labels_checks:
            name_checks.extend([c.raise_context() for c in label_checks])
        name_checks = agg_checks(name_checks)

        # -- DNA checks --

        for check_g, check_l, check_n in DNA_CHECKS:
            for label_i, label_analysis in enumerate(labels_analysis):
                if label_analysis is not None:
                    for grapheme_i, grapheme in enumerate(label_analysis.graphemes):
                        labels_graphemes_checks[label_i][grapheme_i].append(check_g(grapheme))
                labels_checks[label_i].append(check_l(label_analysis))
            name_checks.append(check_n(labels_analysis))

        # -- generate result --

        return NameGuardReport(
            name=name,
            namehash=namehash_from_name(name),
            normalization=Normalization.UNKNOWN
            if any(label_analysis is None for label_analysis in labels_analysis)
            else Normalization.NORMALIZED
            if all(label_analysis.status == 'normalized' and len(label_analysis.label) > 0
                   for label_analysis in labels_analysis)
            else Normalization.UNNORMALIZED,
            rating=calculate_nameguard_rating(name_checks),
            risk_count=count_risks(name_checks),
            highest_risk=get_highest_risk(name_checks),
            checks=sorted(name_checks, reverse=True),
            canonical_name=compute_canonical_from_list(
                [label_analysis.normalized_canonical_label
                 if label_analysis is not None
                 else labels[i] # labelhash
                 for i, label_analysis in enumerate(labels_analysis)],
                 sep='.',
            ),
            labels=[
                LabelGuardReport(
                    # actual label or [labelhash]
                    label=label,
                    labelhash=labelhash_from_label(
                        label_analysis.label) if label_analysis is not None else '0x' + label[1:-1],
                    normalization=Normalization.UNKNOWN
                    if label_analysis is None
                    else Normalization.NORMALIZED
                    if label_analysis.status == 'normalized' and len(label_analysis.label) > 0
                    else Normalization.UNNORMALIZED,
                    rating=calculate_nameguard_rating(label_checks),
                    risk_count=count_risks(label_checks),
                    highest_risk=get_highest_risk(label_checks),
                    checks=sorted(label_checks, reverse=True),
                    canonical_label=label_analysis.normalized_canonical_label if label_analysis is not None else label, # labelhash
                    graphemes=[
                        ConsolidatedGraphemeGuardReport(
                            normalization=GraphemeNormalization.NORMALIZED
                            if any(check.status == CheckStatus.PASS and check.check is Check.NORMALIZED
                                   for check in grapheme_checks)
                            else GraphemeNormalization.UNNORMALIZED,
                            grapheme=grapheme.value,
                            grapheme_name=capitalize_words(grapheme.name),
                            grapheme_type=grapheme.type,
                            grapheme_script=grapheme.script,
                            grapheme_link=grapheme.link,
                            rating=calculate_nameguard_rating(grapheme_checks),
                            risk_count=count_risks(grapheme_checks),
                            highest_risk=get_highest_risk(grapheme_checks),
                            grapheme_description=grapheme.description,
                        )
                        for grapheme, grapheme_checks in zip(label_analysis.graphemes, label_graphemes_checks)
                    ] if label_analysis is not None else None,
                )
                for label, label_analysis, label_checks, label_graphemes_checks in zip(
                    labels,
                    labels_analysis,
                    labels_checks,
                    labels_graphemes_checks,
                )
            ],
        )

    async def bulk_inspect_names(self, network_name: NetworkName, names: list[str]) -> BulkNameGuardBulkReport:
        names = await resolve_all_labelhashes_in_names_querying_labelhashes(network_name, names)
        return BulkNameGuardBulkReport(
            results=[await self.inspect_name(network_name, name, resolve_labelhashes=False) for name in names],
        )

    async def inspect_namehash(self, network_name: NetworkName, namehash: str) -> NameGuardReport:
        logger.debug(f'[inspect_namehash] namehash: \'{namehash}\'')
        name = await namehash_to_name_lookup(network_name, namehash)
        return await self.inspect_name(network_name, name)

    async def inspect_name_with_labelhash_lookup(self, network_name: NetworkName, name: str) -> NameGuardReport:
        '''
        Inspect a name. A name is a sequence of labels separated by dots.
        A label can be a labelhash or a string.
        If a labelhash is encountered, the entire name will be looked up by its namehash.
        If the namehash is not found, all labelhashes will be treated as unknown labels.
        '''

        logger.debug(f'[inspect_name_with_labelhash_lookup] name: \'{name}\'')

        return await self.inspect_name(network_name, name)

    def inspect_grapheme(self, grapheme: str) -> GraphemeGuardReport:
        '''
        Inspect a single grapheme.
        Throws `NotAGrapheme` if the input is not a single grapheme.
        '''

        label_analysis = self.analyse_label(grapheme)
        if label_analysis.grapheme_length != 1:
            raise NotAGrapheme(f'The input contains {label_analysis.grapheme_length} graphemes.')

        grapheme_analysis = label_analysis.graphemes[0]
        grapheme_checks = [check(grapheme_analysis) for check in GRAPHEME_CHECKS + [c[0] for c in DNA_CHECKS]]

        return GraphemeGuardReport(
            normalization=GraphemeNormalization.NORMALIZED
            if any(check.status == CheckStatus.PASS and check.check is Check.NORMALIZED
                   for check in grapheme_checks)
            else GraphemeNormalization.UNNORMALIZED,
            grapheme=grapheme_analysis.value,
            grapheme_name=capitalize_words(grapheme_analysis.name),
            grapheme_type=grapheme_analysis.type,
            grapheme_script=grapheme_analysis.script,
            grapheme_link=grapheme_analysis.link,
            rating=calculate_nameguard_rating(grapheme_checks),
            risk_count=count_risks(grapheme_checks),
            highest_risk=get_highest_risk(grapheme_checks),
            checks=sorted(grapheme_checks, reverse=True),
            confusables=([self._inspect_confusable(grapheme_analysis.confusables_canonical)]
                                 if grapheme_analysis.confusables_canonical else []) + 
                        ([self._inspect_confusable(c)
                         for c in grapheme_analysis.confusables_other]
                         if grapheme_analysis.confusables_other else []),
            canonical_grapheme=label_analysis.canonical_label,
            grapheme_description=grapheme_analysis.description,
        )

    def _inspect_confusable(self, grapheme: InspectorConfusableGraphemeResult) -> ConsolidatedGraphemeGuardReport:
        grapheme_checks = [check(grapheme) for check in GRAPHEME_CHECKS + [c[0] for c in DNA_CHECKS]]
        return ConsolidatedGraphemeGuardReport(
            normalization=GraphemeNormalization.NORMALIZED
            if any(check.status == CheckStatus.PASS and check.check is Check.NORMALIZED
                   for check in grapheme_checks)
            else GraphemeNormalization.UNNORMALIZED,
            grapheme=grapheme.value,
            grapheme_name=capitalize_words(grapheme.name),
            grapheme_type=grapheme.type,
            grapheme_script=grapheme.script,
            grapheme_link=grapheme.link,
            rating=calculate_nameguard_rating(grapheme_checks),
            risk_count=count_risks(grapheme_checks),
            highest_risk=get_highest_risk(grapheme_checks),
            grapheme_description=grapheme.description,
        )

    async def secure_primary_name(self, address: str, network_name: str) -> SecurePrimaryNameResult:
        try:
            domain = self.ns[network_name].name(address)
        except requests.exceptions.ConnectionError as ex:
            raise ProviderUnavailable(f"Communication error with provider occurred: {ex}")
        display_name = f'Unnamed {address[2:6].lower()}'
        primary_name = None
        nameguard_result = None
        if domain is None:
            status = SecurePrimaryNameStatus.NO_PRIMARY_NAME
            impersonation_status = None
        else:
            nameguard_result = await self.inspect_name(network_name, domain)
            
            result = ens_process(domain, do_normalize=True, do_beautify=True)
            if result.normalized != domain:
                status = SecurePrimaryNameStatus.UNNORMALIZED
                impersonation_status = None
            else:
                display_name = result.beautified
                status = SecurePrimaryNameStatus.NORMALIZED
                primary_name = domain

                impersonation_status = ImpersonationStatus.UNLIKELY if any(check.check == 'impersonation_risk' and check.status == CheckStatus.PASS for check in
                    nameguard_result.checks) else ImpersonationStatus.POTENTIAL


        return SecurePrimaryNameResult(primary_name=primary_name,
                                       impersonation_status=impersonation_status,
                                       display_name=display_name,
                                       primary_name_status=status,
                                       nameguard_result=nameguard_result)

    async def fake_eth_name_check_fields(self, network_name, contract_address, token_id, investigated_fields: dict[str,str]) -> FakeEthNameCheckResult:
        contract_address = contract_address.lower()

        return await self._fake_eth_name_check(network_name, contract_address, investigated_fields)

    async def fake_eth_name_check(self, network_name, contract_address, token_id) -> FakeEthNameCheckResult:
        """
        Check if the token is a fake ENS name (not valid ENS contract address and title and collection name of NFT look like ENS name).
        """
        contract_address = contract_address.lower()

        
        res_json = await get_nft_metadata(contract_address, token_id)

        token_type = res_json['id']['tokenMetadata']['tokenType']
        
        if token_type not in ['ERC721', 'ERC1155'] and contract_address in ens_contract_adresses:
            return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_result=None, investigated_fields=None)
        if token_type == 'NOT_A_CONTRACT':
            return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_result=None, investigated_fields=None)
        elif token_type == 'NO_SUPPORTED_NFT_STANDARD':
            return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_result=None, investigated_fields=None)
        elif token_type not in ['ERC721', 'ERC1155']:  # Alchemy does not support other types
            return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_result=None, investigated_fields=None)

        title = res_json['title']
        investigated_fields = {}
        for keys in [['title'],
                     ['metadata', 'name'],
                     ['contractMetadata', 'openSea', 'collectionName'],
                     ['contractMetadata', 'name']]:
            try:
                name = nested_get(res_json, keys)
                investigated_fields['.'.join(keys)] = name
            except KeyError:
                pass

        if contract_address in ens_contract_adresses:
            if ALCHEMY_UNKNOWN_NAME.match(title):
                unknown_name = f"[{res_json['id']['tokenId'][2:]}].eth"
                investigated_fields['title'] = unknown_name
            elif title == '': # the name has never been registered
                investigated_fields['title'] = None
        
       
        return await self._fake_eth_name_check(network_name, contract_address, investigated_fields)
        
    async def _fake_eth_name_check(self, network_name, contract_address, fields: dict[str,str]) -> FakeEthNameCheckResult:

        if contract_address in ens_contract_adresses:
            if 'title' not in fields:
                raise MissingTitle()
            
            title = fields['title']
            
            if title is None:  
                return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_NFT, nameguard_result=None,
                                              investigated_fields=None)
            else:
                if is_labelhash_eth(title):
                    report = await self.inspect_name(network_name, title, resolve_labelhashes=False)
                    return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.UNKNOWN_ETH_NAME,
                                                  nameguard_result=report, investigated_fields=None)

                # TODO: check if token_id matches contract_address and title
                report = await self.inspect_name(network_name, title)
                if is_ens_normalized(title):
                    return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.AUTHENTIC_ETH_NAME,
                                                  nameguard_result=report, investigated_fields=None)
                else:
                    return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.INVALID_ETH_NAME,
                                                  nameguard_result=report, investigated_fields=None)
        else:
            impersonating_fields = {}
            impersonated = False
            potentially_impersonated = False
            for key, name in fields.items():
                try:
                    cured_title = ens_cure(name)  # TODO improve, e.g. remove invisible and then canonicalize
                    inspector_result = self.analyse_label(cured_title)
                    if inspector_result.canonical_label is not None:
                        canonical_name = inspector_result.canonical_label
                    else:
                        canonical_name = cured_title
                except DisallowedSequence:
                    canonical_name = name
                
                if canonical_name.endswith('.eth'):
                    impersonating_fields[key] = name
                    impersonated = True
                elif '.eth' in canonical_name:
                    impersonating_fields[key] = name
                    potentially_impersonated = True
                    
            if impersonated:
                return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.IMPERSONATED_ETH_NAME,
                                                  nameguard_result=None, investigated_fields=impersonating_fields)
            elif potentially_impersonated:
                return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.POTENTIALLY_IMPERSONATED_ETH_NAME,
                                              nameguard_result=None, investigated_fields=impersonating_fields)
            else:
                 return FakeEthNameCheckResult(status=FakeEthNameCheckStatus.NON_IMPERSONATED_ETH_NAME,
                                          nameguard_result=None, investigated_fields=None)



import json
from enum import Enum
from fastapi import FastAPI, Path, Request
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel, Field

import nameguard.context
from nameguard.nameguard import NameGuard
from nameguard.utils import validate_namehash, namehash_from_labelhash, validate_token_id, validate_ethereum_address
from nameguard.models import (
    NameGuardReport,
    BulkNameGuardBulkReport,
    SecurePrimaryNameResult,
    NetworkName,
    GraphemeGuardReport,
    FakeEthNameCheckResult
)
from nameguard.endpoints import Endpoints
from nameguard.logging import logger
from nameguard.exceptions import (
    InvalidNameHash,
    ENSSubgraphUnavailable,
    NamehashMismatchError,
    NamehashNotFoundInSubgraph,
    InvalidEthereumAddress,
    ProviderUnavailable,
    InvalidTokenID,
    NotAGrapheme,
    MissingTitle,
)


class ApiVersion(str, Enum):
    V08_BETA = 'v0.8-beta'


app = FastAPI(
    title='NameGuard API',
    version=ApiVersion.V08_BETA.value,
    description=
'''Welcome to NameGuard, a powerful tool developed by [NameHash Labs](https://namehashlabs.org) to identify and prevent malicious use of Ethereum Name Service (ENS) names.

## Important Notes
⚠️ Beta Version: Please be aware that NameGuard is currently in beta. We appreciate your feedback as we work towards a stable v1 release. Expect changes to the API to address community feedback and improve functionality.

## Documentation
These documentation pages focus specifically on the NameGuard API. For information on the NameGuard Library, SDK, and UI Kit, please refer to the [NameGuard GitHub repository](https://github.com/namehash/nameguard).''',
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

ng = NameGuard()


# -- inspect-name --


class InspectNameRequest(BaseModel):
    name: str = Field(
        description='Name to inspect.',
        examples=['vitalìk.eth'],
    )
    network_name: NetworkName


@app.get(
    '/inspect-name/{network_name}/{name:path}',
    tags=['name'],
    summary='Inspect Name',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
@app.get('/inspect-name/{network_name}', include_in_schema=False)
async def inspect_name_get(
        
        network_name: NetworkName,
        name: str = Path(default_factory=lambda: '',
                         description='**Name should be url-encoded (except when using the Swagger UI).**',
                         examples=['vitalìk.eth']),
) -> NameGuardReport:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_NAME, 'method': 'GET', 'network_name': network_name, 'name': name})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAME)
    return await ng.inspect_name(network_name, name)


@app.post(
    '/inspect-name',
    tags=['name'],
    summary='Inspect Name',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def inspect_name_post( request: InspectNameRequest) -> NameGuardReport:
    logger.debug(f"{json.dumps({'endpoint': Endpoints.INSPECT_NAME, 'method': 'POST', 'network_name': request.network_name, 'name': request.name})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAME)
    return await ng.inspect_name(request.network_name, request.name)


# -- bulk-inspect-names --


class BulkInspectNamesRequest(BaseModel):
    names: list[str] = Field(max_length=250, examples=[['vitalìk.eth', 'nick.eth']])
    network_name: NetworkName


@app.post(
    '/bulk-inspect-names',
    tags=['name'],
    summary='Inspect Multiple Names',
    responses={
        **ENSSubgraphUnavailable.get_responses_spec(),
    },
)
async def bulk_inspect_names( request: BulkInspectNamesRequest) -> BulkNameGuardBulkReport:
    logger.debug(f"{json.dumps({'endpoint': Endpoints.BULK_INSPECT_NAMES, 'method': 'POST', 'network_name': request.network_name, 'names': request.names})}")
    nameguard.context.endpoint_name.set(Endpoints.BULK_INSPECT_NAMES)
    return await ng.bulk_inspect_names(request.network_name, request.names)


# -- inspect-namehash --


class InspectNamehashRequest(BaseModel):
    namehash: str = Field(description='Namehash should be a decimal or a hex (prefixed with 0x) string.',
                          examples=['0xd48fd5598e605861cbd8e45419b41b83739bff52eaef0e283181bbe0a43a5b32'])
    network_name: NetworkName


@app.get(
    '/inspect-namehash/{network_name}/{namehash}',
    tags=['namehash'],
    summary='Inspect Namehash GET',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_namehash_get(
        
        network_name: NetworkName,
        namehash: str = Path(examples=['0xd48fd5598e605861cbd8e45419b41b83739bff52eaef0e283181bbe0a43a5b32'],
                             description='Namehash should be a decimal or a hex (prefixed with 0x) string.')
) -> NameGuardReport:
    logger.debug(f"{json.dumps({'endpoint': Endpoints.INSPECT_NAMEHASH, 'method': 'GET', 'network_name': network_name, 'namehash': namehash})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAMEHASH)
    return await ng.inspect_namehash(network_name=network_name,
                                     namehash=validate_namehash(namehash=namehash),
                                     )


@app.post(
    '/inspect-namehash',
    tags=['namehash'],
    summary='Inspect Namehash',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_namehash_post( request: InspectNamehashRequest) -> NameGuardReport:
    logger.debug(f"{json.dumps({'endpoint': Endpoints.INSPECT_NAMEHASH, 'method': 'POST', 'network_name': request.network_name, 'namehash': request.namehash})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_NAMEHASH)
    return await ng.inspect_namehash(network_name=request.network_name,
                                     namehash=validate_namehash(namehash=request.namehash),
                                     )


# -- inspect-labelhash --


class InspectLabelhashRequest(BaseModel):
    labelhash: str = Field(description='Labelhash should be a decimal or a hex (prefixed with 0x) string.',
                           examples=['0x3276e4878615389906712b876ce1455b8f5d1c5ea3ffcf7a705e0d32fafae9c5'])
    network_name: NetworkName
    parent_name: str = Field('eth')


@app.get(
    '/inspect-labelhash/{network_name}/{labelhash}/{parent_name:path}',
    tags=['labelhash'],
    summary='Inspect Labelhash GET',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
@app.get('/inspect-labelhash/{network_name}/{labelhash}', include_in_schema=False)
async def inspect_labelhash_get(
        
        network_name: NetworkName,
        labelhash: str = Path(examples=['0x3276e4878615389906712b876ce1455b8f5d1c5ea3ffcf7a705e0d32fafae9c5'],
                              description='Labelhash should be a decimal or a hex (prefixed with 0x) string.'),
        parent_name: str = Path(default_factory=lambda: '', examples=['eth'])
) -> NameGuardReport:
    logger.debug(f"{json.dumps({'endpoint': Endpoints.INSPECT_LABELHASH, 'method': 'GET', 'network_name': network_name, 'labelhash': labelhash, 'parent_name': parent_name})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_LABELHASH)
    valid_labelhash = validate_namehash(namehash=labelhash)
    namehash = namehash_from_labelhash(valid_labelhash, parent_name=parent_name)
    return await ng.inspect_namehash(network_name=network_name, namehash=namehash)


@app.post(
    '/inspect-labelhash',
    tags=['labelhash'],
    summary='Inspect Labelhash',
    responses={
        **InvalidNameHash.get_responses_spec(),
        **ENSSubgraphUnavailable.get_responses_spec(),
        **NamehashMismatchError.get_responses_spec(),
        **NamehashNotFoundInSubgraph.get_responses_spec(),
    },
)
async def inspect_labelhash_post( request: InspectLabelhashRequest) -> NameGuardReport:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_LABELHASH, 'method': 'POST', 'network_name': request.network_name, 'labelhash': request.labelhash, 'parent_name': request.parent_name})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_LABELHASH)
    valid_labelhash = validate_namehash(namehash=request.labelhash)
    namehash = namehash_from_labelhash(valid_labelhash, parent_name=request.parent_name)
    return await ng.inspect_namehash(network_name=request.network_name, namehash=namehash)


# -- secure-primary-name --


@app.get(
    '/secure-primary-name/{network_name}/{address}',
    tags=['secure_primary_name'],
    summary='Reverse lookup of Ethereum address to primary name',
    responses={
        **InvalidEthereumAddress.get_responses_spec(),
        **ProviderUnavailable.get_responses_spec(),
    },
)
async def secure_primary_name_get( address: str,
                                  network_name: NetworkName) -> SecurePrimaryNameResult:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.SECURE_PRIMARY_NAME, 'method': 'GET', 'network_name': network_name, 'address': address})}")
    nameguard.context.endpoint_name.set(Endpoints.SECURE_PRIMARY_NAME)
    address = validate_ethereum_address(address)
    return await ng.secure_primary_name(address, network_name)


# -- fake-ens-name-check --


@app.get(
    '/fake-eth-name-check/{network_name}/{contract_address}/{token_id}',
    tags=['fake-eth-name-check'],
    summary='Fake .eth ENS name check GET',
    responses={
        **InvalidTokenID.get_responses_spec(),
        **InvalidEthereumAddress.get_responses_spec(),
        **ProviderUnavailable.get_responses_spec(),
    },
)
async def fake_eth_name_check_get(
        
        network_name: NetworkName,
        contract_address: str = Path(examples=['0x495f947276749ce646f68ac8c248420045cb7b5e'],
                                     description='Contract address for the NFT contract (ERC721 and ERC1155 supported).'),
        token_id: str = Path(examples=['61995921128521442959106650131462633744885269624153038309795231243542768648193'],
                             description='The ID of the token (in hex or decimal format).')
) -> FakeEthNameCheckResult:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.FAKE_ETH_NAME_CHECK, 'method': 'GET', 'network_name': network_name, 'contract_address': contract_address, 'token_id': token_id})}")
    nameguard.context.endpoint_name.set(Endpoints.FAKE_ETH_NAME_CHECK)
    contract_address = validate_ethereum_address(contract_address)
    token_id = validate_token_id(token_id)
    return await ng.fake_eth_name_check(network_name=network_name, contract_address=contract_address, token_id=token_id)

class FakeETHNameCheckFieldsRequest(BaseModel):
    network_name: NetworkName
    contract_address: str = Path(examples=['0x495f947276749ce646f68ac8c248420045cb7b5e'],
                                 description='Contract address for the NFT contract (ERC721 and ERC1155 supported).')
    token_id: str = Path(examples=['61995921128521442959106650131462633744885269624153038309795231243542768648193'],
                         description='The ID of the token (in hex or decimal format).')
    fields: dict[str, str] = Field(
        description='Fields with values which will be investigated (e.g. title, collection name, metadata) whether they look like fake .eth ENS name. `title` key is mandatory, for ENS contracts it should be the ENS name.')


@app.post(
    '/fake-eth-name-check-fields',
    tags=['fake-eth-name-check'],
    summary='Fake .eth ENS name check with fields',
    responses={
        **InvalidTokenID.get_responses_spec(),
        **InvalidEthereumAddress.get_responses_spec(),
        **MissingTitle.get_responses_spec(),
    },
)
async def fake_eth_name_check_fields_post(
         
        request: FakeETHNameCheckFieldsRequest
        ) -> FakeEthNameCheckResult:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.FAKE_ETH_NAME_CHECK, 'method': 'POST', 'network_name': request.network_name, 'contract_address': request.contract_address, 'token_id': request.token_id})}")
    nameguard.context.endpoint_name.set(Endpoints.FAKE_ETH_NAME_CHECK)
    contract_address = validate_ethereum_address(request.contract_address)
    token_id = validate_token_id(request.token_id)
    return await ng.fake_eth_name_check_fields(network_name=request.network_name, contract_address=contract_address, token_id=token_id, investigated_fields=request.fields)



# -- inspect-grapheme --


@app.get(
    '/inspect-grapheme/{grapheme}',
    tags=['grapheme'],
    summary='Inspect Grapheme GET',
    responses={
        **NotAGrapheme.get_responses_spec(),
    },
)
async def inspect_grapheme_get(
        grapheme: str = Path(
            description='Grapheme to inspect. Should be url-encoded (except when using the Swagger UI).',
            examples=['v', 'ń', '%F0%9F%98%B5'])
) -> GraphemeGuardReport:
    logger.debug(
        f"{json.dumps({'endpoint': Endpoints.INSPECT_GRAPHEME, 'method': 'GET', 'grapheme': grapheme})}")
    nameguard.context.endpoint_name.set(Endpoints.INSPECT_GRAPHEME)
    return ng.inspect_grapheme(grapheme)


if __name__ == '__main__':
    import asyncio

    asyncio.run(ng.inspect_name('mainnet', 'nick.eth'))



from label_inspector.models import InspectorGraphemeWithConfusablesResult as Grapheme
from nameguard.models import CheckStatus, Check, GenericCheckResult, GraphemeCheckResult


STATUS = CheckStatus.WARN

MESSAGE_PASS = 'Probably supported by common fonts'

MESSAGE_FAIL = 'May not be supported by common fonts'

MESSAGE_SKIP = 'Unknown if supported by common fonts'


def check_grapheme(grapheme: Grapheme) -> GenericCheckResult:
    passed = grapheme.font_support_all_os
    if passed is None:
        return GraphemeCheckResult(
            check=Check.FONT_SUPPORT,
            status=CheckStatus.SKIP,
            _grapheme_message=MESSAGE_SKIP,
            _label_message=MESSAGE_SKIP,
            _name_message=MESSAGE_SKIP,
        )
    else:
        return GraphemeCheckResult(
            check=Check.FONT_SUPPORT,
            status=CheckStatus.PASS if passed else STATUS,
            _grapheme_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
            _label_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
            _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
        )



from typing import Optional
from nameguard.models import CheckStatus, Check, GenericCheckResult, NameCheckResult
from label_inspector.common.punycode import puny_analysis, PunycodeCompatibility
from label_inspector.models import InspectorResult


STATUS = CheckStatus.WARN
MESSAGE_PASS = 'Compatible for use with DNS'
MESSAGE_FAIL = 'Incompatible for use with DNS'
MESSAGE_SKIP = 'This name contains unknown labels'


def check_name(labels: list[Optional[InspectorResult]]) -> GenericCheckResult:
    if any(label is None for label in labels):
        return NameCheckResult(
            check=Check.PUNYCODE_COMPATIBLE_NAME,
            status=CheckStatus.SKIP,
            _name_message=MESSAGE_SKIP,
        )
    result = puny_analysis('.'.join(label.label for label in labels))
    passed = result.compatibility == PunycodeCompatibility.COMPATIBLE
    return NameCheckResult(
        check=Check.PUNYCODE_COMPATIBLE_NAME,
        status=CheckStatus.PASS if passed else STATUS,
        _name_message=MESSAGE_PASS if passed else MESSAGE_FAIL,
    )



from enum import Enum


class Endpoints(str, Enum):
    INSPECT_NAME = 'inspect-name'
    INSPECT_NAMEHASH = 'inspect-namehash'
    INSPECT_LABELHASH = 'inspect-labelhash'
    INSPECT_GRAPHEME = 'inspect-grapheme'
    BULK_INSPECT_NAMES = 'bulk-inspect-names'
    SECURE_PRIMARY_NAME = 'secure-primary-name'
    FAKE_ETH_NAME_CHECK = 'fake-eth-name-check'



import os
import httpx

from nameguard.logging import logger
from nameguard.exceptions import ProviderUnavailable


async def get_nft_metadata(contract_address: str, token_id: str) -> dict:
    url = f"{os.environ.get('PROVIDER_URI_MAINNET')}/getNFTMetadata?contractAddress={contract_address}&tokenId={token_id}&refreshCache=false"
    headers = {"accept": "application/json"}

    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                url,
                headers=headers,
            )

        if response.status_code == 200:
            response_json = response.json()
            logger.debug(f"Provider response json:\n{response_json}")
            if 'error' in response_json and 'timed out' in response_json['error']:
                raise ProviderUnavailable(
                    f"Received retryable error from provider {response.status_code}: {response_json['error']}")
        else:
            raise ProviderUnavailable(
                f"Received unexpected status code from provider {response.status_code}: {response.text}")
    except Exception as ex:
        logger.exception(f'Communication error with provider occurred')
        if not str(ex):
            raise ProviderUnavailable(f"Communication error with provider occurred.")
        raise ProviderUnavailable(f"Communication error with provider occurred: {ex}")

    return response_json
